/* $Id: test-tcpproxy.cc,v 1.2 2009/01/25 20:01:32 tudorm Exp $ */

/*
 * tcpproxy tester by Robert Morris
 *
 * g++ -g -I/usr/local/include/sfs -I/usr/local/include tprox.C -L/usr/local/lib -L/usr/local/lib/sfs -lasync -ldmalloc -o tprox
 *
 */

#include "stdio.h"
#include "stdlib.h"
#include "amisc.h"
#include <sys/types.h>
#include <sys/socket.h>
#include "async.h"

void
accept_cb(int s, callback<void, int>::ref cb)
{
  int s1;
  unsigned sinlen;
  struct sockaddr_in sin;

  sinlen = sizeof(sin);
  s1 = accept(s, (struct sockaddr *) &sin, &sinlen);
  assert(s1 >= 0);
  make_async(s1);
  tcp_nodelay(s1);

  cb(s1);
}

/*
 * Create a TCP server. Returns port the server listens to.
 * cb will be called for each newly accepted connection.
 */
int
make_server(callback<void, int>::ref cb)
{
  int s;
  unsigned sinlen;
  struct sockaddr_in sin;

  s = inetsocket(SOCK_STREAM, 0, INADDR_ANY);
  assert(s >= 0);
  make_async(s);
  tcp_nodelay(s);
  listen(s, 128);

  sinlen = sizeof(sin);
  assert(getsockname(s, (struct sockaddr *) &sin, &sinlen) == 0);

  fdcb(s, selread, wrap(accept_cb, s, cb));

  return(sin.sin_port);
}

void
write_n_cb(int s, char *buf, int off, int len, callback<void>::ref cb)
{
  int cc;

  cc = write(s, buf + off, len);
  if(cc >= len){
    fdcb(s, selwrite, 0);
    free(buf);
    cb();
  } else if(cc > 0){
    fdcb(s, selwrite, wrap(write_n_cb, s, buf, off + cc, len - cc, cb));
  } else {
    fdcb(s, selwrite, 0);
    free(buf);
    cb();
  }
}

void
write_n(int s, char *buf, int n, callback<void>::ref cb)
{
  char *p = (char *) malloc(n);
  assert(p);
  bcopy(buf, p, n);
  fdcb(s, selwrite, wrap(write_n_cb, s, p, 0, n, cb));
}

void
read_n_cb(int s, char *p, int off, int n, callback<void, char *, int>::ref cb)
{
  int cc;

  fdcb(s, selread, 0);

  cc = read(s, p + off, n - off);
  if(cc <= 0 || off + cc >= n){
    cb(p, off + cc);
  } else {
    fdcb(s, selread, wrap(read_n_cb, s, p, off + cc, n, cb));
  }
}

void
read_n(int s, int n, callback<void, char *, int>::ref cb)
{
  char *p = (char *) malloc(n);
  assert(p);
  fdcb(s, selread, wrap(read_n_cb, s, p, 0, n, cb));
}

void echo_server_2(int, int);

void
echo_read(int rs, int ws)
{
  char buf[4096];
  int n;

  fdcb(rs, selread, 0);

  n = read(rs, buf, sizeof(buf));
  if(n > 0){
    write_n(ws, buf, n, wrap(echo_server_2, rs, ws));
  } else {
    close(ws);
    close(rs);
  }
}

void
echo_server_2(int rs, int ws)
{
  fdcb(rs, selread, wrap(echo_read, rs, ws));
}

void
xxx_cb(int s)
{
  int cc;
  char buf[512];

  fdcb(s, selread, 0);

  cc = read(s, buf, sizeof(buf));
  if(cc <= 0){
    write(s, "z", 1);
    close(s);
  } else {
    fdcb(s, selread, wrap(xxx_cb, s));
  }
}

void
any_cb(int s, char *p, int pn)
{
  int mode;

  if(p && pn == sizeof(int))
    bcopy(p, &mode, sizeof(int));
  if(pn == 4 && mode == 0){
    // ordinary echo
    echo_server_2(s, s);
  } else if(pn == 4 && mode == 1){
    // ignore
  } else if(pn == 4 && mode == 2){
    shutdown (s, SHUT_WR);
    delaycb (0, 250000000, wrap (tcp_abort, s));
  } else if(pn == 4 && mode == 3){
    // wait for EOF, then write "z"
    fdcb(s, selread, wrap(xxx_cb, s));
  } else {
    fprintf(stderr, "tprox: bad server mode %d\n", mode);
    close(s);
  }
}

void
any_server(int s)
{
  read_n(s, sizeof(int), wrap(any_cb, s));
}

void echoc_write(int, int, int, callback<void, int>::ref cb);

void
echoc_read_cb(int s, int seq, int n, callback<void, int>::ref cb, char *p, int pn)
{
  int x;

  if(pn == 4){
    bcopy(p, &x, 4);
    if(x != seq){
      close(s);
      cb(0);
    } else if(seq >= n){
      close(s);
      cb(1);
    } else {
      echoc_write(s, seq + 1, n, cb);
    }
  } else {
    close(s);
    cb(0);
  }
  free(p);
}

void
echoc_read(int s, int seq, int n, callback<void, int>::ref cb)
{
  read_n(s, sizeof(int), wrap(echoc_read_cb, s, seq, n, cb));;
}

void
echoc_write_cb(int s, int seq, int n, callback<void, int>::ref cb)
{
  echoc_read(s, seq, n, cb);
}

void
echoc_write(int s, int seq, int n, callback<void, int>::ref cb)
{
  write_n(s, (char*)&seq, sizeof(seq), wrap(echoc_write_cb, s, seq, n, cb));
}

void
echo_client(int n, callback<void, int>::ref cb, int s)
{
  echoc_write(s, 1, n, cb);
}

void
mc_cb_cb(callback<void, int>::ref cb, int s)
{
  cb(s);
}

void done();

void
mc_cb(int mode, callback<void, int>::ref cb, int s)
{
  if(s < 0){
    printf("tprox: connect to proxy failed; cannot continue.\n");
    done();
  }
  write_n(s, (char *)&mode, sizeof(mode), wrap(mc_cb_cb, cb, s));
}

// mode is
//   0: ordinary echo
void
make_client(int port, int mode, callback<void, int>::ref cb)
{
  tcpconnect("127.0.0.1", ntohs(port), wrap(mc_cb, mode, cb));
}

class info {
public:
  info() { wanted = good = bad = 0; }
  info(int w) { wanted = w; good = bad = 0; }
  int wanted;
  int good;
  int bad;
};

void
chain2(int port, int n, callback<void, int, int>::ref cb, int s0,
       int sa, int sb)
{
  if(n <= 0 && sa == s0){
    echo_server_2(dup(sa), dup(sb));
    cb(sa, sb);
  } else if(n <= 0){
    echo_server_2(sa, dup(sb));
    cb(s0, sb);
  } else if(sa == s0){
    echo_server_2(dup(sa), dup(sb));
    make_client(port, 0, wrap(chain2, port, n - 1, cb, s0, sb));
  } else {
    echo_server_2(sa, dup(sb));
    make_client(port, 0, wrap(chain2, port, n - 1, cb, s0, sb));
  }
}

void
chain1(int port, int n, callback<void, int, int>::ref cb, int s)
{
  make_client(port, 0, wrap(chain2, port, n - 1, cb, s, s));
}

void
make_chain(int port, int n, callback<void, int, int>::ref cb)
{
  make_client(port, 0, wrap(chain1, port, n - 1, cb));
}

void
ec_d(callback<void, int>::ref cb, int s1, int s2, int x, char *p, int pn)
{
  int y;

  if(pn == sizeof(y))
    bcopy(p, &y, sizeof(y));

  if(pn == sizeof(y) && y == x){
    cb(1);
  } else {
    cb(0);
  }

  free(p);
  shutdown(s1, SHUT_WR);
}

void
ec_c(callback<void, int>::ref cb, int s1, int s2, int x)
{
  read_n(s2, sizeof(int), wrap(ec_d, cb, s1, s2, x));
}

void
ec_b(callback<void, int>::ref cb, int s1, int s2)
{
  int x = random();
  write_n(s1, (char*)&x, sizeof(x), wrap(ec_c, cb, s1, s2, x));
}

void
echo_chain(int port, int n, callback<void, int>::ref cb)
{
  make_chain(port, n, wrap(ec_b, cb));
}

void
nothing()
{
}

void
bc_c(callback<void, int>::ref cb, int s1, int s2, int len, char *p, int pn)
{
  int ok = 1;
  int *ip;
  for(ip = (int *)p; ip < (int *)(p + pn); ip++)
    if(*ip != (ip - (int *)p) + len)
      ok = 0;

  free(p);
  shutdown(s1, SHUT_WR);

  if(pn == len && ok){
    cb(1);
  } else {
    cb(0);
  }
}

void
bc_bb(callback<void, int>::ref cb, int s1, int s2, int len)
{
  read_n(s2, len, wrap(bc_c, cb, s1, s2, len));
}

void
bc_b(callback<void, int>::ref cb, int s1, int s2)
{
  int len = 2000000 + 4*(random() % 10000);
  char *p = (char *) malloc(len);
  int *ip;
  for(ip = (int *)p; ip < (int *)(p + len); ip++)
    *ip = (ip - (int *)p) + len;
  write_n(s1, p, len, wrap(nothing));
  free(p);
  delaycb(2, 0, wrap(bc_bb, cb, s1, s2, len));
}

void
bulk_chain(int port, callback<void, int>::ref cb)
{
  make_chain(port, 20, wrap(bc_b, cb));
}

void
p8_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  done();
}

void
p8_c(int port, int s, time_t t0, int bytes)
{
  int cc;
  char buf[4096];
  time_t t1;

  time(&t1);
  fdcb(s, selwrite, 0);

  if(t1 - t0 > 21){
    // Proxy never timed out the connection.
    close(s);
    p8_done(port, 0);
    return;
  }

  cc = write(s, buf, sizeof(buf));
  
  if(cc > 0){
    fdcb(s, selwrite, wrap(p8_c, port, s, t0, bytes + cc));
  } else {
    close(s);
    if(t1 - t0 > 8){
      // Proxy closed connection in about ten seconds -- good.
      p8_done(port, 1);
    } else {
      // Proxy closed the connection too soon.
      p8_done(port, 0);
    }
  }
}

void
p8_b(int port, int s)
{
  time_t t;
  time(&t);
  fdcb(s, selwrite, wrap(p8_c, port, s, t, 0));
}

void
start_phase_8(int port)
{
  // If we send infinite data but the reader never reads, does
  // the proxy close the connection in ten seconds?
  printf("Timeout of lazy client: ");
  make_client(port, 1, wrap(p8_b, port));
}

void
p7_d(int port, int s, char *p, int pn)
{
  close(s);
  if(pn == 2 && p[0] == 'x' && p[1] == 'y'){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  free(p);
  start_phase_8(port);
}

void
p7_c(int port, int s)
{
  write(s, "y", 1);
  read_n(s, 2, wrap(p7_d, port, s));
}

void
p7_b(int port, int s)
{
  write(s, "x", 1);
  delaycb(21, 0, wrap(p7_c, port, s));
}

void
start_phase_7(int port)
{
  printf("Non-timeout of active client: ");
  make_client(port, 0, wrap(p7_b, port));
}

void
p6_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  start_phase_7(port);
}

void
p6_b(int port, int n, int s)
{
  int cc = write(s, "x", 1);
  if(n <= 1){
    close(s);
    if(cc == 1){
      p6_done(port, 0);
    } else {
      echo_chain(port, 2, wrap(p6_done, port));
    }
  } else {
    delaycb(1, 0, wrap(p6_b, port, n - 1, s));
  }
}

void
start_phase_6(int port)
{
  // I.e. produce a write error in the proxy.
  printf("Early close: ");
  make_client(port, 2, wrap(p6_b, port, 5));
}

void
p55_c(int port, int s, char *p, int pn)
{
  if(p && pn > 0 && p[0] == 'z'){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  free(p);
  close(s);
  start_phase_6(port);
}

void
p55_b(int port, int s)
{
  shutdown(s, SHUT_WR);
  read_n(s, 1, wrap(p55_c, port, s));
}

void
start_phase_55(int port)
{
  // If we shutdown() for writing, can we still read?
  printf("One-way shutdown: ");
  make_client(port, 3, wrap(p55_b, port));
}

void
p5_block_cb(int s)
{
  char buf[1024];
  int cc;
  
  fdcb(s, selwrite, 0);

  cc = write(s, buf, sizeof(buf));
  if(cc <= 0){
    close(s);
  } else {
    fdcb(s, selwrite, wrap(p5_block_cb, s));
  }
}

void
p5_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  start_phase_55(port);
}

void
p5_b(int port, int s)
{
  fdcb(s, selwrite, wrap(p5_block_cb, s));
  bulk_chain(port, wrap(p5_done, port));
}

void
start_phase_5(int port)
{
  // If one connection is blocked (i.e. somebody isn't reading),
  // can other connections make progress?
  printf("Mix of blocked and normal: ");
  make_client(port, 1, wrap(p5_b, port));
}

void
p4_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  start_phase_5(port);
}

void
start_phase_4(int port)
{
  // Chain 2 megabytes through a sequence of connections.
  printf("Bulk data, 20 connections: ");
  bulk_chain(port, wrap(p4_done, port));
}

void
p3_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  start_phase_4(port);
}

void
start_phase_3(int port)
{
  // Send 4 bytes through a chain of 20 connections.
  printf("20 echo connections: ");
  echo_chain(port, 20, wrap(p3_done, port));
}

void
p2_done(int port, int ok)
{
  if(ok){
    printf("passed\n");
  } else {
    printf("failed\n");
  }
  start_phase_3(port);
}

void
start_phase_2(int port)
{
  printf("Two echo connections: ");
  echo_chain(port, 2, wrap(p2_done, port));
}

void
done_phase_1(ref<info> i, int port, int ok)
{
  if(ok)
    i->good += 1;
  else
    i->bad += 1;

  if(i->good + i->bad >= i->wanted){
    if(i->good == i->wanted && i->bad == 0){
      printf("passed\n");
    } else {
      printf("failed\n");
    }
    start_phase_2(port);
  }
}

void
start_phase_1(int port)
{
  ref<info> i = New refcounted<info> (1);

  printf("Single echo connection: ");

  make_client(port, 0, wrap(echo_client, 10, wrap(done_phase_1, i, port)));
}

void start_child(char *, int, int);

void
child_cb(char *progname, int cport, int tport, int status)
{
  printf("tprox: child tcpproxy failed; cannot continue.\n");
  done();
}

int tprox_pid;

void
start_child(char *progname, int cport, int tport)
{
  int pid;
  char tbuf[64], cbuf[64];

  sprintf(tbuf, "%d", tport); // test program
  sprintf(cbuf, "%d", cport); // child (tcpproxy) program

  pid = fork();
  if(pid < 0){
    perror("tprox: fork");
    exit(1);
  }
  if(pid == 0){
    struct rlimit rl;
    int i;
    for(i = 3; i < 100; i++)
      close(i);
#if 0
    // suppress output from proxy
    int fd = open("/dev/null", 2);
    dup2(fd, 0);
    dup2(fd, 1);
    dup2(fd, 2);
#endif
    getrlimit(RLIMIT_DATA, &rl);
    if(rl.rlim_cur > 100 * 1024 * 1024)
      rl.rlim_cur = 100 * 1024 * 1024;
    setrlimit(RLIMIT_DATA, &rl);
    execl(progname, progname, "127.0.0.1", tbuf, cbuf, (char *) NULL);
    perror("tprox: could not exec tcpproxy program");
    exit(1);
  }
  tprox_pid = pid;
  sleep(2);
  chldcb(pid, wrap(child_cb, progname, cport, tport));
}

void
done()
{
  if(tprox_pid > 0)
    kill(tprox_pid, 15);
  exit(0);
}

int
main(int argc, char **argv)
{
  setprogname (argv[0]);

  char *proxname;
  int p1, p2;

  if(argc != 2){
    warnx << "usage: " << progname << " ./tcpproxy\n";
    exit(1);
  }
  proxname = argv[1];

  srandom(getpid());
  setbuf(stdout, 0);
  p1 = 40000 + (random() % 20000);
  p2 = make_server(wrap(any_server));

  start_child(proxname, p1, ntohs(p2));

  start_phase_1(htons(p1));

  amain();
}

